"""Utilities used by insteon component."""
import asyncio
import logging

from pyinsteon import devices
from pyinsteon.constants import ALDBStatus
from pyinsteon.events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from pyinsteon.managers.link_manager import (
    async_enter_linking_mode,
    async_enter_unlinking_mode,
)
from pyinsteon.managers.scene_manager import (
    async_trigger_scene_off,
    async_trigger_scene_on,
)
from pyinsteon.managers.x10_manager import (
    async_x10_all_lights_off,
    async_x10_all_lights_on,
    async_x10_all_units_off,
)

from homeassistant.const import CONF_ADDRESS, CONF_ENTITY_ID, ENTITY_MATCH_ALL
from homeassistant.core import callback
from homeassistant.helpers import discovery
from homeassistant.helpers.dispatcher import (
    async_dispatcher_connect,
    async_dispatcher_send,
    dispatcher_send,
)

from .const import (
    DOMAIN,
    EVENT_CONF_BUTTON,
    EVENT_GROUP_OFF,
    EVENT_GROUP_OFF_FAST,
    EVENT_GROUP_ON,
    EVENT_GROUP_ON_FAST,
    ON_OFF_EVENTS,
    SIGNAL_ADD_DEFAULT_LINKS,
    SIGNAL_LOAD_ALDB,
    SIGNAL_PRINT_ALDB,
    SIGNAL_SAVE_DEVICES,
    SRV_ADD_ALL_LINK,
    SRV_ADD_DEFAULT_LINKS,
    SRV_ALL_LINK_GROUP,
    SRV_ALL_LINK_MODE,
    SRV_CONTROLLER,
    SRV_DEL_ALL_LINK,
    SRV_HOUSECODE,
    SRV_LOAD_ALDB,
    SRV_LOAD_DB_RELOAD,
    SRV_PRINT_ALDB,
    SRV_PRINT_IM_ALDB,
    SRV_SCENE_OFF,
    SRV_SCENE_ON,
    SRV_X10_ALL_LIGHTS_OFF,
    SRV_X10_ALL_LIGHTS_ON,
    SRV_X10_ALL_UNITS_OFF,
)
from .ipdb import get_device_platforms, get_platform_groups
from .schemas import (
    ADD_ALL_LINK_SCHEMA,
    ADD_DEFAULT_LINKS_SCHEMA,
    DEL_ALL_LINK_SCHEMA,
    LOAD_ALDB_SCHEMA,
    PRINT_ALDB_SCHEMA,
    TRIGGER_SCENE_SCHEMA,
    X10_HOUSECODE_SCHEMA,
)

_LOGGER = logging.getLogger(__name__)


def add_on_off_event_device(hass, device):
    """Register an Insteon device as an on/off event device."""

    @callback
    def async_fire_group_on_off_event(name, address, group, button):
        # Firing an event when a button is pressed.
        if button and button[-2] == "_":
            button_id = button[-1].lower()
        else:
            button_id = None

        schema = {CONF_ADDRESS: address}
        if button_id:
            schema[EVENT_CONF_BUTTON] = button_id
        if name == ON_EVENT:
            event = EVENT_GROUP_ON
        if name == OFF_EVENT:
            event = EVENT_GROUP_OFF
        if name == ON_FAST_EVENT:
            event = EVENT_GROUP_ON_FAST
        if name == OFF_FAST_EVENT:
            event = EVENT_GROUP_OFF_FAST
        _LOGGER.debug("Firing event %s with %s", event, schema)
        hass.bus.async_fire(event, schema)

    for group in device.events:
        if isinstance(group, int):
            for event in device.events[group]:
                if event in [
                    OFF_EVENT,
                    ON_EVENT,
                    OFF_FAST_EVENT,
                    ON_FAST_EVENT,
                ]:
                    _LOGGER.debug(
                        "Registering on/off event for %s %d %s",
                        str(device.address),
                        group,
                        event,
                    )
                    device.events[group][event].subscribe(
                        async_fire_group_on_off_event, force_strong_ref=True
                    )


def register_new_device_callback(hass, config):
    """Register callback for new Insteon device."""
    new_device_lock = asyncio.Lock()

    @callback
    def async_new_insteon_device(address=None):
        """Detect device from transport to be delegated to platform."""
        hass.async_create_task(async_create_new_entities(address))

    async def async_create_new_entities(address):
        _LOGGER.debug(
            "Adding new INSTEON device to Home Assistant with address %s", address
        )
        async with new_device_lock:
            await devices.async_save(workdir=hass.config.config_dir)
        device = devices[address]
        await device.async_status()
        platforms = get_device_platforms(device)
        tasks = []
        for platform in platforms:
            if platform == ON_OFF_EVENTS:
                add_on_off_event_device(hass, device)

            else:
                tasks.append(
                    discovery.async_load_platform(
                        hass,
                        platform,
                        DOMAIN,
                        discovered={"address": device.address.id},
                        hass_config=config,
                    )
                )
        await asyncio.gather(*tasks)

    devices.subscribe(async_new_insteon_device, force_strong_ref=True)


@callback
def async_register_services(hass):
    """Register services used by insteon component."""

    async def async_srv_add_all_link(service):
        """Add an INSTEON All-Link between two devices."""
        group = service.data.get(SRV_ALL_LINK_GROUP)
        mode = service.data.get(SRV_ALL_LINK_MODE)
        link_mode = mode.lower() == SRV_CONTROLLER
        await async_enter_linking_mode(link_mode, group)

    async def async_srv_del_all_link(service):
        """Delete an INSTEON All-Link between two devices."""
        group = service.data.get(SRV_ALL_LINK_GROUP)
        await async_enter_unlinking_mode(group)

    async def async_srv_load_aldb(service):
        """Load the device All-Link database."""
        entity_id = service.data[CONF_ENTITY_ID]
        reload = service.data[SRV_LOAD_DB_RELOAD]
        if entity_id.lower() == ENTITY_MATCH_ALL:
            await async_srv_load_aldb_all(reload)
        else:
            signal = f"{entity_id}_{SIGNAL_LOAD_ALDB}"
            async_dispatcher_send(hass, signal, reload)

    async def async_srv_load_aldb_all(reload):
        """Load the All-Link database for all devices."""
        # Cannot be done concurrently due to issues with the underlying protocol.
        for address in devices:
            device = devices[address]
            if device != devices.modem and device.cat != 0x03:
                await device.aldb.async_load(
                    refresh=reload, callback=async_srv_save_devices
                )

    async def async_srv_save_devices():
        """Write the Insteon device configuration to file."""
        _LOGGER.debug("Saving Insteon devices")
        await devices.async_save(hass.config.config_dir)

    def print_aldb(service):
        """Print the All-Link Database for a device."""
        # For now this sends logs to the log file.
        # Future direction is to create an INSTEON control panel.
        entity_id = service.data[CONF_ENTITY_ID]
        signal = f"{entity_id}_{SIGNAL_PRINT_ALDB}"
        dispatcher_send(hass, signal)

    def print_im_aldb(service):
        """Print the All-Link Database for a device."""
        # For now this sends logs to the log file.
        # Future direction is to create an INSTEON control panel.
        print_aldb_to_log(devices.modem.aldb)

    async def async_srv_x10_all_units_off(service):
        """Send the X10 All Units Off command."""
        housecode = service.data.get(SRV_HOUSECODE)
        await async_x10_all_units_off(housecode)

    async def async_srv_x10_all_lights_off(service):
        """Send the X10 All Lights Off command."""
        housecode = service.data.get(SRV_HOUSECODE)
        await async_x10_all_lights_off(housecode)

    async def async_srv_x10_all_lights_on(service):
        """Send the X10 All Lights On command."""
        housecode = service.data.get(SRV_HOUSECODE)
        await async_x10_all_lights_on(housecode)

    async def async_srv_scene_on(service):
        """Trigger an INSTEON scene ON."""
        group = service.data.get(SRV_ALL_LINK_GROUP)
        await async_trigger_scene_on(group)

    async def async_srv_scene_off(service):
        """Trigger an INSTEON scene ON."""
        group = service.data.get(SRV_ALL_LINK_GROUP)
        await async_trigger_scene_off(group)

    @callback
    def async_add_default_links(service):
        """Add the default All-Link entries to a device."""
        entity_id = service.data[CONF_ENTITY_ID]
        signal = f"{entity_id}_{SIGNAL_ADD_DEFAULT_LINKS}"
        async_dispatcher_send(hass, signal)

    hass.services.async_register(
        DOMAIN, SRV_ADD_ALL_LINK, async_srv_add_all_link, schema=ADD_ALL_LINK_SCHEMA
    )
    hass.services.async_register(
        DOMAIN, SRV_DEL_ALL_LINK, async_srv_del_all_link, schema=DEL_ALL_LINK_SCHEMA
    )
    hass.services.async_register(
        DOMAIN, SRV_LOAD_ALDB, async_srv_load_aldb, schema=LOAD_ALDB_SCHEMA
    )
    hass.services.async_register(
        DOMAIN, SRV_PRINT_ALDB, print_aldb, schema=PRINT_ALDB_SCHEMA
    )
    hass.services.async_register(DOMAIN, SRV_PRINT_IM_ALDB, print_im_aldb, schema=None)
    hass.services.async_register(
        DOMAIN,
        SRV_X10_ALL_UNITS_OFF,
        async_srv_x10_all_units_off,
        schema=X10_HOUSECODE_SCHEMA,
    )
    hass.services.async_register(
        DOMAIN,
        SRV_X10_ALL_LIGHTS_OFF,
        async_srv_x10_all_lights_off,
        schema=X10_HOUSECODE_SCHEMA,
    )
    hass.services.async_register(
        DOMAIN,
        SRV_X10_ALL_LIGHTS_ON,
        async_srv_x10_all_lights_on,
        schema=X10_HOUSECODE_SCHEMA,
    )
    hass.services.async_register(
        DOMAIN, SRV_SCENE_ON, async_srv_scene_on, schema=TRIGGER_SCENE_SCHEMA
    )
    hass.services.async_register(
        DOMAIN, SRV_SCENE_OFF, async_srv_scene_off, schema=TRIGGER_SCENE_SCHEMA
    )

    hass.services.async_register(
        DOMAIN,
        SRV_ADD_DEFAULT_LINKS,
        async_add_default_links,
        schema=ADD_DEFAULT_LINKS_SCHEMA,
    )
    async_dispatcher_connect(hass, SIGNAL_SAVE_DEVICES, async_srv_save_devices)
    _LOGGER.debug("Insteon Services registered")


def print_aldb_to_log(aldb):
    """Print the All-Link Database to the log file."""
    # This service is useless if the log level is not INFO for the
    # insteon component. Setting the log level to INFO and resetting it
    # back when we are done
    orig_log_level = _LOGGER.level
    if orig_log_level > logging.INFO:
        _LOGGER.setLevel(logging.INFO)
    _LOGGER.info("%s ALDB load status is %s", aldb.address, aldb.status.name)
    if aldb.status not in [ALDBStatus.LOADED, ALDBStatus.PARTIAL]:
        _LOGGER.warning("All-Link database not loaded")

    _LOGGER.info("RecID In Use Mode HWM Group Address  Data 1 Data 2 Data 3")
    _LOGGER.info("----- ------ ---- --- ----- -------- ------ ------ ------")
    for mem_addr in aldb:
        rec = aldb[mem_addr]
        # For now we write this to the log
        # Roadmap is to create a configuration panel
        in_use = "Y" if rec.is_in_use else "N"
        mode = "C" if rec.is_controller else "R"
        hwm = "Y" if rec.is_high_water_mark else "N"
        log_msg = (
            f" {rec.mem_addr:04x}    {in_use:s}     {mode:s}   {hwm:s}    "
            f"{rec.group:3d} {str(rec.target):s}   {rec.data1:3d}   "
            f"{rec.data2:3d}   {rec.data3:3d}"
        )
        _LOGGER.info(log_msg)
    _LOGGER.setLevel(orig_log_level)


@callback
def async_add_insteon_entities(
    hass, platform, entity_type, async_add_entities, discovery_info
):
    """Add Insteon devices to a platform."""
    new_entities = []
    device_list = [discovery_info.get("address")] if discovery_info else devices

    for address in device_list:
        device = devices[address]
        groups = get_platform_groups(device, platform)
        for group in groups:
            new_entities.append(entity_type(device, group))
    if new_entities:
        async_add_entities(new_entities)



"""Support for Telegram bot using polling."""
import logging

from telegram import Update
from telegram.error import NetworkError, RetryAfter, TelegramError, TimedOut
from telegram.ext import Handler, Updater

from homeassistant.const import EVENT_HOMEASSISTANT_START, EVENT_HOMEASSISTANT_STOP
from homeassistant.core import callback

from . import CONF_ALLOWED_CHAT_IDS, BaseTelegramBotEntity, initialize_bot

_LOGGER = logging.getLogger(__name__)


async def async_setup_platform(hass, config):
    """Set up the Telegram polling platform."""
    bot = initialize_bot(config)
    pol = TelegramPoll(bot, hass, config[CONF_ALLOWED_CHAT_IDS])

    @callback
    def _start_bot(_event):
        """Start the bot."""
        pol.start_polling()

    @callback
    def _stop_bot(_event):
        """Stop the bot."""
        pol.stop_polling()

    hass.bus.async_listen_once(EVENT_HOMEASSISTANT_START, _start_bot)
    hass.bus.async_listen_once(EVENT_HOMEASSISTANT_STOP, _stop_bot)

    return True


def process_error(bot, update, error):
    """Telegram bot error handler."""
    try:
        raise error
    except (TimedOut, NetworkError, RetryAfter):
        # Long polling timeout or connection problem. Nothing serious.
        pass
    except TelegramError:
        _LOGGER.error('Update "%s" caused error "%s"', update, error)


def message_handler(handler):
    """Create messages handler."""

    class MessageHandler(Handler):
        """Telegram bot message handler."""

        def __init__(self):
            """Initialize the messages handler instance."""
            super().__init__(handler)

        def check_update(self, update):
            """Check is update valid."""
            return isinstance(update, Update)

        def handle_update(self, update, dispatcher):
            """Handle update."""
            optional_args = self.collect_optional_args(dispatcher, update)
            return self.callback(dispatcher.bot, update, **optional_args)

    return MessageHandler()


class TelegramPoll(BaseTelegramBotEntity):
    """Asyncio telegram incoming message handler."""

    def __init__(self, bot, hass, allowed_chat_ids):
        """Initialize the polling instance."""

        BaseTelegramBotEntity.__init__(self, hass, allowed_chat_ids)

        self.updater = Updater(bot=bot, workers=4)
        self.dispatcher = self.updater.dispatcher

        self.dispatcher.add_handler(message_handler(self.process_update))
        self.dispatcher.add_error_handler(process_error)

    def start_polling(self):
        """Start the polling task."""
        self.updater.start_polling()

    def stop_polling(self):
        """Stop the polling task."""
        self.updater.stop()

    def process_update(self, bot, update):
        """Process incoming message."""
        self.process_message(update.to_dict())



"""Support for N26 bank account sensors."""
import logging

from homeassistant.helpers.entity import Entity

from . import DEFAULT_SCAN_INTERVAL, DOMAIN, timestamp_ms_to_date
from .const import DATA

_LOGGER = logging.getLogger(__name__)

SCAN_INTERVAL = DEFAULT_SCAN_INTERVAL

ATTR_IBAN = "account"
ATTR_USABLE_BALANCE = "usable_balance"
ATTR_BANK_BALANCE = "bank_balance"

ATTR_ACC_OWNER_TITLE = "owner_title"
ATTR_ACC_OWNER_FIRST_NAME = "owner_first_name"
ATTR_ACC_OWNER_LAST_NAME = "owner_last_name"
ATTR_ACC_OWNER_GENDER = "owner_gender"
ATTR_ACC_OWNER_BIRTH_DATE = "owner_birth_date"
ATTR_ACC_OWNER_EMAIL = "owner_email"
ATTR_ACC_OWNER_PHONE_NUMBER = "owner_phone_number"

ICON_ACCOUNT = "mdi:currency-eur"
ICON_CARD = "mdi:credit-card"
ICON_SPACE = "mdi:crop-square"


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the N26 sensor platform."""
    if discovery_info is None:
        return

    api_list = hass.data[DOMAIN][DATA]

    sensor_entities = []
    for api_data in api_list:
        sensor_entities.append(N26Account(api_data))

        for card in api_data.cards:
            sensor_entities.append(N26Card(api_data, card))

        for space in api_data.spaces["spaces"]:
            sensor_entities.append(N26Space(api_data, space))

    add_entities(sensor_entities)


class N26Account(Entity):
    """Sensor for a N26 balance account.

    A balance account contains an amount of money (=balance). The amount may
    also be negative.
    """

    def __init__(self, api_data) -> None:
        """Initialize a N26 balance account."""
        self._data = api_data
        self._iban = self._data.balance["iban"]

    def update(self) -> None:
        """Get the current balance and currency for the account."""
        self._data.update_account()

    @property
    def unique_id(self):
        """Return the unique ID of the entity."""
        return self._iban[-4:]

    @property
    def name(self) -> str:
        """Friendly name of the sensor."""
        return f"n26_{self._iban[-4:]}"

    @property
    def state(self) -> float:
        """Return the balance of the account as state."""
        if self._data.balance is None:
            return None

        return self._data.balance.get("availableBalance")

    @property
    def unit_of_measurement(self) -> str:
        """Use the currency as unit of measurement."""
        if self._data.balance is None:
            return None

        return self._data.balance.get("currency")

    @property
    def device_state_attributes(self) -> dict:
        """Additional attributes of the sensor."""
        attributes = {
            ATTR_IBAN: self._data.balance.get("iban"),
            ATTR_BANK_BALANCE: self._data.balance.get("bankBalance"),
            ATTR_USABLE_BALANCE: self._data.balance.get("usableBalance"),
            ATTR_ACC_OWNER_TITLE: self._data.account_info.get("title"),
            ATTR_ACC_OWNER_FIRST_NAME: self._data.account_info.get("kycFirstName"),
            ATTR_ACC_OWNER_LAST_NAME: self._data.account_info.get("kycLastName"),
            ATTR_ACC_OWNER_GENDER: self._data.account_info.get("gender"),
            ATTR_ACC_OWNER_BIRTH_DATE: timestamp_ms_to_date(
                self._data.account_info.get("birthDate")
            ),
            ATTR_ACC_OWNER_EMAIL: self._data.account_info.get("email"),
            ATTR_ACC_OWNER_PHONE_NUMBER: self._data.account_info.get(
                "mobilePhoneNumber"
            ),
        }

        for limit in self._data.limits:
            limit_attr_name = f"limit_{limit['limit'].lower()}"
            attributes[limit_attr_name] = limit["amount"]

        return attributes

    @property
    def icon(self) -> str:
        """Set the icon for the sensor."""
        return ICON_ACCOUNT


class N26Card(Entity):
    """Sensor for a N26 card."""

    def __init__(self, api_data, card) -> None:
        """Initialize a N26 card."""
        self._data = api_data
        self._account_name = api_data.balance["iban"][-4:]
        self._card = card

    def update(self) -> None:
        """Get the current balance and currency for the account."""
        self._data.update_cards()
        self._card = self._data.card(self._card["id"], self._card)

    @property
    def unique_id(self):
        """Return the unique ID of the entity."""
        return self._card["id"]

    @property
    def name(self) -> str:
        """Friendly name of the sensor."""
        return f"{self._account_name.lower()}_card_{self._card['id']}"

    @property
    def state(self) -> float:
        """Return the balance of the account as state."""
        return self._card["status"]

    @property
    def device_state_attributes(self) -> dict:
        """Additional attributes of the sensor."""
        attributes = {
            "apple_pay_eligible": self._card.get("applePayEligible"),
            "card_activated": timestamp_ms_to_date(self._card.get("cardActivated")),
            "card_product": self._card.get("cardProduct"),
            "card_product_type": self._card.get("cardProductType"),
            "card_settings_id": self._card.get("cardSettingsId"),
            "card_Type": self._card.get("cardType"),
            "design": self._card.get("design"),
            "exceet_actual_delivery_date": self._card.get("exceetActualDeliveryDate"),
            "exceet_card_status": self._card.get("exceetCardStatus"),
            "exceet_expected_delivery_date": self._card.get(
                "exceetExpectedDeliveryDate"
            ),
            "exceet_express_card_delivery": self._card.get("exceetExpressCardDelivery"),
            "exceet_express_card_delivery_email_sent": self._card.get(
                "exceetExpressCardDeliveryEmailSent"
            ),
            "exceet_express_card_delivery_tracking_id": self._card.get(
                "exceetExpressCardDeliveryTrackingId"
            ),
            "expiration_date": timestamp_ms_to_date(self._card.get("expirationDate")),
            "google_pay_eligible": self._card.get("googlePayEligible"),
            "masked_pan": self._card.get("maskedPan"),
            "membership": self._card.get("membership"),
            "mpts_card": self._card.get("mptsCard"),
            "pan": self._card.get("pan"),
            "pin_defined": timestamp_ms_to_date(self._card.get("pinDefined")),
            "username_on_card": self._card.get("usernameOnCard"),
        }
        return attributes

    @property
    def icon(self) -> str:
        """Set the icon for the sensor."""
        return ICON_CARD


class N26Space(Entity):
    """Sensor for a N26 space."""

    def __init__(self, api_data, space) -> None:
        """Initialize a N26 space."""
        self._data = api_data
        self._space = space

    def update(self) -> None:
        """Get the current balance and currency for the account."""
        self._data.update_spaces()
        self._space = self._data.space(self._space["id"], self._space)

    @property
    def unique_id(self):
        """Return the unique ID of the entity."""
        return f"space_{self._data.balance['iban'][-4:]}_{self._space['name'].lower()}"

    @property
    def name(self) -> str:
        """Friendly name of the sensor."""
        return self._space["name"]

    @property
    def state(self) -> float:
        """Return the balance of the account as state."""
        return self._space["balance"]["availableBalance"]

    @property
    def unit_of_measurement(self) -> str:
        """Use the currency as unit of measurement."""
        return self._space["balance"]["currency"]

    @property
    def device_state_attributes(self) -> dict:
        """Additional attributes of the sensor."""
        goal_value = ""
        if "goal" in self._space:
            goal_value = self._space.get("goal").get("amount")

        attributes = {
            "name": self._space.get("name"),
            "goal": goal_value,
            "background_image_url": self._space.get("backgroundImageUrl"),
            "image_url": self._space.get("imageUrl"),
            "is_card_attached": self._space.get("isCardAttached"),
            "is_hidden_from_balance": self._space.get("isHiddenFromBalance"),
            "is_locked": self._space.get("isLocked"),
            "is_primary": self._space.get("isPrimary"),
        }
        return attributes

    @property
    def icon(self) -> str:
        """Set the icon for the sensor."""
        return ICON_SPACE



"""Sensors flow for Withings."""
from typing import Callable, List

from homeassistant.components.binary_sensor import (
    DEVICE_CLASS_PRESENCE,
    DOMAIN as BINARY_SENSOR_DOMAIN,
    BinarySensorDevice,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant
from homeassistant.helpers.entity import Entity

from .common import BaseWithingsSensor, async_create_entities


async def async_setup_entry(
    hass: HomeAssistant,
    entry: ConfigEntry,
    async_add_entities: Callable[[List[Entity], bool], None],
) -> None:
    """Set up the sensor config entry."""
    entities = await async_create_entities(
        hass, entry, WithingsHealthBinarySensor, BINARY_SENSOR_DOMAIN
    )

    async_add_entities(entities, True)


class WithingsHealthBinarySensor(BaseWithingsSensor, BinarySensorDevice):
    """Implementation of a Withings sensor."""

    @property
    def is_on(self) -> bool:
        """Return true if the binary sensor is on."""
        return self._state_data

    @property
    def device_class(self) -> str:
        """Provide the device class."""
        return DEVICE_CLASS_PRESENCE



"""Binary sensors for the Elexa Guardian integration."""
from homeassistant.components.binary_sensor import BinarySensorEntity
from homeassistant.core import callback

from . import GuardianEntity
from .const import (
    DATA_CLIENT,
    DATA_SENSOR_STATUS,
    DATA_WIFI_STATUS,
    DOMAIN,
    SENSOR_KIND_AP_INFO,
    SENSOR_KIND_LEAK_DETECTED,
)

ATTR_CONNECTED_CLIENTS = "connected_clients"

SENSORS = [
    (SENSOR_KIND_AP_INFO, "Onboard AP Enabled", "connectivity"),
    (SENSOR_KIND_LEAK_DETECTED, "Leak Detected", "moisture"),
]


async def async_setup_entry(hass, entry, async_add_entities):
    """Set up Guardian switches based on a config entry."""
    guardian = hass.data[DOMAIN][DATA_CLIENT][entry.entry_id]
    async_add_entities(
        [
            GuardianBinarySensor(guardian, kind, name, device_class)
            for kind, name, device_class in SENSORS
        ],
        True,
    )


class GuardianBinarySensor(GuardianEntity, BinarySensorEntity):
    """Define a generic Guardian sensor."""

    def __init__(self, guardian, kind, name, device_class):
        """Initialize."""
        super().__init__(guardian, kind, name, device_class, None)

        self._is_on = True

    @property
    def is_on(self):
        """Return True if the binary sensor is on."""
        return self._is_on

    @callback
    def _update_from_latest_data(self):
        """Update the entity."""
        if self._kind == SENSOR_KIND_AP_INFO:
            self._is_on = self._guardian.data[DATA_WIFI_STATUS]["ap_enabled"]
            self._attrs.update(
                {
                    ATTR_CONNECTED_CLIENTS: self._guardian.data[DATA_WIFI_STATUS][
                        "ap_clients"
                    ]
                }
            )
        elif self._kind == SENSOR_KIND_LEAK_DETECTED:
            self._is_on = self._guardian.data[DATA_SENSOR_STATUS]["wet"]



"""Config flow for Gogogate2."""
import logging
import re

from gogogate2_api.common import ApiError
from gogogate2_api.const import ApiErrorCode
import voluptuous as vol

from homeassistant import config_entries
from homeassistant.config_entries import SOURCE_IMPORT, ConfigFlow
from homeassistant.const import CONF_IP_ADDRESS, CONF_PASSWORD, CONF_USERNAME

from .common import get_api
from .const import DOMAIN  # pylint: disable=unused-import

_LOGGER = logging.getLogger(__name__)


class Gogogate2FlowHandler(ConfigFlow, domain=DOMAIN):
    """Gogogate2 config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_POLL

    async def async_step_import(self, config_data: dict = None):
        """Handle importing of configuration."""
        result = await self.async_step_user(config_data)
        self._abort_if_unique_id_configured()
        return result

    async def async_step_user(self, user_input: dict = None):
        """Handle user initiated flow."""
        user_input = user_input or {}
        errors = {}

        if user_input:
            api = get_api(user_input)
            try:
                data = await self.hass.async_add_executor_job(api.info)
                await self.async_set_unique_id(re.sub("\\..*$", "", data.remoteaccess))
                return self.async_create_entry(title=data.gogogatename, data=user_input)

            except ApiError as api_error:
                if api_error.code in (
                    ApiErrorCode.CREDENTIALS_NOT_SET,
                    ApiErrorCode.CREDENTIALS_INCORRECT,
                ):
                    errors["base"] = "invalid_auth"
                else:
                    errors["base"] = "cannot_connect"

            except Exception:  # pylint: disable=broad-except
                errors["base"] = "cannot_connect"

        if errors and self.source == SOURCE_IMPORT:
            return self.async_abort(reason="cannot_connect")

        return self.async_show_form(
            step_id="user",
            data_schema=vol.Schema(
                {
                    vol.Required(
                        CONF_IP_ADDRESS, default=user_input.get(CONF_IP_ADDRESS, "")
                    ): str,
                    vol.Required(
                        CONF_USERNAME, default=user_input.get(CONF_USERNAME, "admin")
                    ): str,
                    vol.Required(
                        CONF_PASSWORD, default=user_input.get(CONF_PASSWORD, "")
                    ): str,
                }
            ),
            errors=errors,
        )



"""Config flow for Speedtest.net."""
import logging

import voluptuous as vol

from homeassistant import config_entries
from homeassistant.const import CONF_MONITORED_CONDITIONS, CONF_SCAN_INTERVAL
from homeassistant.core import callback

from . import server_id_valid
from .const import (
    CONF_MANUAL,
    CONF_SERVER_ID,
    CONF_SERVER_NAME,
    DEFAULT_NAME,
    DEFAULT_SCAN_INTERVAL,
    DEFAULT_SERVER,
)
from .const import DOMAIN  # pylint: disable=unused-import

_LOGGER = logging.getLogger(__name__)


class SpeedTestFlowHandler(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle Speedtest.net config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_POLL

    @staticmethod
    @callback
    def async_get_options_flow(config_entry):
        """Get the options flow for this handler."""
        return SpeedTestOptionsFlowHandler(config_entry)

    async def async_step_user(self, user_input=None):
        """Handle a flow initialized by the user."""
        if self._async_current_entries():
            return self.async_abort(reason="one_instance_allowed")

        if user_input is None:
            return self.async_show_form(step_id="user")

        return self.async_create_entry(title=DEFAULT_NAME, data=user_input)

    async def async_step_import(self, import_config):
        """Import from config."""
        if (
            CONF_SERVER_ID in import_config
            and not await self.hass.async_add_executor_job(
                server_id_valid, import_config[CONF_SERVER_ID]
            )
        ):
            return self.async_abort(reason="wrong_server_id")

        import_config[CONF_SCAN_INTERVAL] = int(
            import_config[CONF_SCAN_INTERVAL].seconds / 60
        )
        import_config.pop(CONF_MONITORED_CONDITIONS)

        return await self.async_step_user(user_input=import_config)


class SpeedTestOptionsFlowHandler(config_entries.OptionsFlow):
    """Handle SpeedTest options."""

    def __init__(self, config_entry):
        """Initialize options flow."""
        self.config_entry = config_entry
        self._servers = {}

    async def async_step_init(self, user_input=None):
        """Manage the options."""
        errors = {}

        if user_input is not None:
            server_name = user_input[CONF_SERVER_NAME]
            if server_name != "*Auto Detect":
                server_id = self._servers[server_name]["id"]
                user_input[CONF_SERVER_ID] = server_id
            else:
                user_input[CONF_SERVER_ID] = None

            return self.async_create_entry(title="", data=user_input)

        self._servers = self.hass.data[DOMAIN].servers

        server_name = DEFAULT_SERVER
        if self.config_entry.options.get(
            CONF_SERVER_ID
        ) and not self.config_entry.options.get(CONF_SERVER_NAME):
            server = [
                key
                for (key, value) in self._servers.items()
                if value.get("id") == self.config_entry.options[CONF_SERVER_ID]
            ]
            server_name = server[0] if server else ""

        options = {
            vol.Optional(
                CONF_SERVER_NAME,
                default=self.config_entry.options.get(CONF_SERVER_NAME, server_name),
            ): vol.In(self._servers.keys()),
            vol.Optional(
                CONF_SCAN_INTERVAL,
                default=self.config_entry.options.get(
                    CONF_SCAN_INTERVAL, DEFAULT_SCAN_INTERVAL
                ),
            ): int,
            vol.Optional(
                CONF_MANUAL, default=self.config_entry.options.get(CONF_MANUAL, False)
            ): bool,
        }

        return self.async_show_form(
            step_id="init", data_schema=vol.Schema(options), errors=errors
        )



"""The Mikrotik component."""
import voluptuous as vol

from homeassistant.config_entries import SOURCE_IMPORT
from homeassistant.const import (
    CONF_HOST,
    CONF_NAME,
    CONF_PASSWORD,
    CONF_PORT,
    CONF_USERNAME,
    CONF_VERIFY_SSL,
)
from homeassistant.helpers import config_validation as cv

from .const import (
    ATTR_MANUFACTURER,
    CONF_ARP_PING,
    CONF_DETECTION_TIME,
    CONF_FORCE_DHCP,
    DEFAULT_API_PORT,
    DEFAULT_DETECTION_TIME,
    DEFAULT_NAME,
    DOMAIN,
)
from .hub import MikrotikHub

MIKROTIK_SCHEMA = vol.All(
    vol.Schema(
        {
            vol.Required(CONF_HOST): cv.string,
            vol.Required(CONF_USERNAME): cv.string,
            vol.Required(CONF_PASSWORD): cv.string,
            vol.Optional(CONF_NAME, default=DEFAULT_NAME): cv.string,
            vol.Optional(CONF_PORT, default=DEFAULT_API_PORT): cv.port,
            vol.Optional(CONF_VERIFY_SSL, default=False): cv.boolean,
            vol.Optional(CONF_ARP_PING, default=False): cv.boolean,
            vol.Optional(CONF_FORCE_DHCP, default=False): cv.boolean,
            vol.Optional(
                CONF_DETECTION_TIME, default=DEFAULT_DETECTION_TIME
            ): cv.time_period,
        }
    )
)

CONFIG_SCHEMA = vol.Schema(
    {DOMAIN: vol.All(cv.ensure_list, [MIKROTIK_SCHEMA])}, extra=vol.ALLOW_EXTRA
)


async def async_setup(hass, config):
    """Import the Mikrotik component from config."""

    if DOMAIN in config:
        for entry in config[DOMAIN]:
            hass.async_create_task(
                hass.config_entries.flow.async_init(
                    DOMAIN, context={"source": SOURCE_IMPORT}, data=entry
                )
            )

    return True


async def async_setup_entry(hass, config_entry):
    """Set up the Mikrotik component."""

    hub = MikrotikHub(hass, config_entry)
    if not await hub.async_setup():
        return False

    hass.data.setdefault(DOMAIN, {})[config_entry.entry_id] = hub
    device_registry = await hass.helpers.device_registry.async_get_registry()
    device_registry.async_get_or_create(
        config_entry_id=config_entry.entry_id,
        connections={(DOMAIN, hub.serial_num)},
        manufacturer=ATTR_MANUFACTURER,
        model=hub.model,
        name=hub.hostname,
        sw_version=hub.firmware,
    )

    return True


async def async_unload_entry(hass, config_entry):
    """Unload a config entry."""
    await hass.config_entries.async_forward_entry_unload(config_entry, "device_tracker")

    hass.data[DOMAIN].pop(config_entry.entry_id)

    return True



"""Support for Synology DSM sensors."""
from homeassistant.config_entries import ConfigEntry
from homeassistant.const import (
    CONF_DISKS,
    DATA_MEGABYTES,
    DATA_RATE_KILOBYTES_PER_SECOND,
    DATA_TERABYTES,
    PRECISION_TENTHS,
    TEMP_CELSIUS,
)
from homeassistant.helpers.temperature import display_temp
from homeassistant.helpers.typing import HomeAssistantType

from . import SynologyDSMDeviceEntity, SynologyDSMEntity
from .const import (
    CONF_VOLUMES,
    DOMAIN,
    STORAGE_DISK_SENSORS,
    STORAGE_VOL_SENSORS,
    SYNO_API,
    TEMP_SENSORS_KEYS,
    UTILISATION_SENSORS,
)


async def async_setup_entry(
    hass: HomeAssistantType, entry: ConfigEntry, async_add_entities
) -> None:
    """Set up the Synology NAS Sensor."""

    api = hass.data[DOMAIN][entry.unique_id][SYNO_API]

    entities = [
        SynoDSMUtilSensor(api, sensor_type, UTILISATION_SENSORS[sensor_type])
        for sensor_type in UTILISATION_SENSORS
    ]

    # Handle all volumes
    if api.storage.volumes_ids:
        for volume in entry.data.get(CONF_VOLUMES, api.storage.volumes_ids):
            entities += [
                SynoDSMStorageSensor(
                    api, sensor_type, STORAGE_VOL_SENSORS[sensor_type], volume
                )
                for sensor_type in STORAGE_VOL_SENSORS
            ]

    # Handle all disks
    if api.storage.disks_ids:
        for disk in entry.data.get(CONF_DISKS, api.storage.disks_ids):
            entities += [
                SynoDSMStorageSensor(
                    api, sensor_type, STORAGE_DISK_SENSORS[sensor_type], disk
                )
                for sensor_type in STORAGE_DISK_SENSORS
            ]

    async_add_entities(entities)


class SynoDSMUtilSensor(SynologyDSMEntity):
    """Representation a Synology Utilisation sensor."""

    @property
    def state(self):
        """Return the state."""
        attr = getattr(self._api.utilisation, self.entity_type)
        if callable(attr):
            attr = attr()
        if attr is None:
            return None

        # Data (RAM)
        if self._unit == DATA_MEGABYTES:
            return round(attr / 1024.0 ** 2, 1)

        # Network
        if self._unit == DATA_RATE_KILOBYTES_PER_SECOND:
            return round(attr / 1024.0, 1)

        return attr

    @property
    def available(self) -> bool:
        """Return True if entity is available."""
        return bool(self._api.utilisation)


class SynoDSMStorageSensor(SynologyDSMDeviceEntity):
    """Representation a Synology Storage sensor."""

    @property
    def state(self):
        """Return the state."""
        attr = getattr(self._api.storage, self.entity_type)(self._device_id)
        if attr is None:
            return None

        # Data (disk space)
        if self._unit == DATA_TERABYTES:
            return round(attr / 1024.0 ** 4, 2)

        # Temperature
        if self.entity_type in TEMP_SENSORS_KEYS:
            return display_temp(self.hass, attr, TEMP_CELSIUS, PRECISION_TENTHS)

        return attr



"""Support for Somfy Camera Shutter."""
from pymfy.api.devices.camera_protect import CameraProtect
from pymfy.api.devices.category import Category

from homeassistant.components.switch import SwitchEntity

from . import API, DEVICES, DOMAIN, SomfyEntity


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up the Somfy switch platform."""

    def get_shutters():
        """Retrieve switches."""
        devices = hass.data[DOMAIN][DEVICES]

        return [
            SomfyCameraShutter(device, hass.data[DOMAIN][API])
            for device in devices
            if Category.CAMERA.value in device.categories
        ]

    async_add_entities(await hass.async_add_executor_job(get_shutters), True)


class SomfyCameraShutter(SomfyEntity, SwitchEntity):
    """Representation of a Somfy Camera Shutter device."""

    def __init__(self, device, api):
        """Initialize the Somfy device."""
        super().__init__(device, api)
        self.shutter = CameraProtect(self.device, self.api)

    async def async_update(self):
        """Update the device with the latest data."""
        await super().async_update()
        self.shutter = CameraProtect(self.device, self.api)

    def turn_on(self, **kwargs) -> None:
        """Turn the entity on."""
        self.shutter.open_shutter()

    def turn_off(self, **kwargs):
        """Turn the entity off."""
        self.shutter.close_shutter()

    @property
    def is_on(self) -> bool:
        """Return True if entity is on."""
        return self.shutter.get_shutter_position() == "opened"



"""Config flow to configure Blink."""
import logging

from blinkpy.blinkpy import Blink
import voluptuous as vol

from homeassistant import config_entries, core, exceptions
from homeassistant.const import (
    CONF_PASSWORD,
    CONF_PIN,
    CONF_SCAN_INTERVAL,
    CONF_USERNAME,
)
from homeassistant.core import callback

from .const import DEFAULT_OFFSET, DEFAULT_SCAN_INTERVAL, DEVICE_ID, DOMAIN

_LOGGER = logging.getLogger(__name__)


async def validate_input(hass: core.HomeAssistant, blink):
    """Validate the user input allows us to connect."""
    response = await hass.async_add_executor_job(blink.get_auth_token)
    if not response:
        raise InvalidAuth
    if blink.key_required:
        raise Require2FA

    return blink.login_response


class BlinkConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a Blink config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_CLOUD_POLL

    def __init__(self):
        """Initialize the blink flow."""
        self.blink = None
        self.data = {
            CONF_USERNAME: "",
            CONF_PASSWORD: "",
            "login_response": None,
        }

    @staticmethod
    @callback
    def async_get_options_flow(config_entry):
        """Get options flow for this handler."""
        return BlinkOptionsFlowHandler(config_entry)

    async def async_step_user(self, user_input=None):
        """Handle a flow initiated by the user."""
        errors = {}
        if user_input is not None:
            self.data[CONF_USERNAME] = user_input["username"]
            self.data[CONF_PASSWORD] = user_input["password"]

            await self.async_set_unique_id(self.data[CONF_USERNAME])

            if CONF_SCAN_INTERVAL in user_input:
                self.data[CONF_SCAN_INTERVAL] = user_input[CONF_SCAN_INTERVAL]

            self.blink = Blink(
                username=self.data[CONF_USERNAME],
                password=self.data[CONF_PASSWORD],
                motion_interval=DEFAULT_OFFSET,
                legacy_subdomain=False,
                no_prompt=True,
                device_id=DEVICE_ID,
            )

            try:
                response = await validate_input(self.hass, self.blink)
                self.data["login_response"] = response
                return self.async_create_entry(title=DOMAIN, data=self.data,)
            except Require2FA:
                return await self.async_step_2fa()
            except InvalidAuth:
                errors["base"] = "invalid_auth"
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Unexpected exception")
                errors["base"] = "unknown"

        data_schema = {
            vol.Required("username"): str,
            vol.Required("password"): str,
        }

        return self.async_show_form(
            step_id="user", data_schema=vol.Schema(data_schema), errors=errors,
        )

    async def async_step_2fa(self, user_input=None):
        """Handle 2FA step."""
        if user_input is not None:
            pin = user_input.get(CONF_PIN)
            if await self.hass.async_add_executor_job(
                self.blink.login_handler.send_auth_key, self.blink, pin
            ):
                return await self.async_step_user(user_input=self.data)

        return self.async_show_form(
            step_id="2fa",
            data_schema=vol.Schema(
                {vol.Optional("pin"): vol.All(str, vol.Length(min=1))}
            ),
        )

    async def async_step_import(self, import_data):
        """Import blink config from configuration.yaml."""
        return await self.async_step_user(import_data)


class BlinkOptionsFlowHandler(config_entries.OptionsFlow):
    """Handle Blink options."""

    def __init__(self, config_entry):
        """Initialize Blink options flow."""
        self.config_entry = config_entry
        self.options = dict(config_entry.options)
        self.blink = None

    async def async_step_init(self, user_input=None):
        """Manage the Blink options."""
        self.blink = self.hass.data[DOMAIN][self.config_entry.entry_id]
        self.options[CONF_SCAN_INTERVAL] = self.blink.refresh_rate

        return await self.async_step_simple_options()

    async def async_step_simple_options(self, user_input=None):
        """For simple options."""
        if user_input is not None:
            self.options.update(user_input)
            self.blink.refresh_rate = user_input[CONF_SCAN_INTERVAL]
            return self.async_create_entry(title="", data=self.options)

        options = self.config_entry.options
        scan_interval = options.get(CONF_SCAN_INTERVAL, DEFAULT_SCAN_INTERVAL)

        return self.async_show_form(
            step_id="simple_options",
            data_schema=vol.Schema(
                {vol.Optional(CONF_SCAN_INTERVAL, default=scan_interval,): int}
            ),
        )


class Require2FA(exceptions.HomeAssistantError):
    """Error to indicate we require 2FA."""


class InvalidAuth(exceptions.HomeAssistantError):
    """Error to indicate there is invalid auth."""



"""Support for Vallox ventilation unit sensors."""

from datetime import datetime, timedelta
import logging

from homeassistant.const import (
    DEVICE_CLASS_HUMIDITY,
    DEVICE_CLASS_TEMPERATURE,
    DEVICE_CLASS_TIMESTAMP,
    TEMP_CELSIUS,
    UNIT_PERCENTAGE,
)
from homeassistant.core import callback
from homeassistant.helpers.dispatcher import async_dispatcher_connect
from homeassistant.helpers.entity import Entity

from . import DOMAIN, METRIC_KEY_MODE, SIGNAL_VALLOX_STATE_UPDATE

_LOGGER = logging.getLogger(__name__)


async def async_setup_platform(hass, config, async_add_entities, discovery_info=None):
    """Set up the sensors."""
    if discovery_info is None:
        return

    name = hass.data[DOMAIN]["name"]
    state_proxy = hass.data[DOMAIN]["state_proxy"]

    sensors = [
        ValloxProfileSensor(
            name=f"{name} Current Profile",
            state_proxy=state_proxy,
            device_class=None,
            unit_of_measurement=None,
            icon="mdi:gauge",
        ),
        ValloxFanSpeedSensor(
            name=f"{name} Fan Speed",
            state_proxy=state_proxy,
            metric_key="A_CYC_FAN_SPEED",
            device_class=None,
            unit_of_measurement=UNIT_PERCENTAGE,
            icon="mdi:fan",
        ),
        ValloxSensor(
            name=f"{name} Extract Air",
            state_proxy=state_proxy,
            metric_key="A_CYC_TEMP_EXTRACT_AIR",
            device_class=DEVICE_CLASS_TEMPERATURE,
            unit_of_measurement=TEMP_CELSIUS,
            icon=None,
        ),
        ValloxSensor(
            name=f"{name} Exhaust Air",
            state_proxy=state_proxy,
            metric_key="A_CYC_TEMP_EXHAUST_AIR",
            device_class=DEVICE_CLASS_TEMPERATURE,
            unit_of_measurement=TEMP_CELSIUS,
            icon=None,
        ),
        ValloxSensor(
            name=f"{name} Outdoor Air",
            state_proxy=state_proxy,
            metric_key="A_CYC_TEMP_OUTDOOR_AIR",
            device_class=DEVICE_CLASS_TEMPERATURE,
            unit_of_measurement=TEMP_CELSIUS,
            icon=None,
        ),
        ValloxSensor(
            name=f"{name} Supply Air",
            state_proxy=state_proxy,
            metric_key="A_CYC_TEMP_SUPPLY_AIR",
            device_class=DEVICE_CLASS_TEMPERATURE,
            unit_of_measurement=TEMP_CELSIUS,
            icon=None,
        ),
        ValloxSensor(
            name=f"{name} Humidity",
            state_proxy=state_proxy,
            metric_key="A_CYC_RH_VALUE",
            device_class=DEVICE_CLASS_HUMIDITY,
            unit_of_measurement=UNIT_PERCENTAGE,
            icon=None,
        ),
        ValloxFilterRemainingSensor(
            name=f"{name} Remaining Time For Filter",
            state_proxy=state_proxy,
            metric_key="A_CYC_REMAINING_TIME_FOR_FILTER",
            device_class=DEVICE_CLASS_TIMESTAMP,
            unit_of_measurement=None,
            icon="mdi:filter",
        ),
    ]

    async_add_entities(sensors, update_before_add=False)


class ValloxSensor(Entity):
    """Representation of a Vallox sensor."""

    def __init__(
        self, name, state_proxy, metric_key, device_class, unit_of_measurement, icon
    ) -> None:
        """Initialize the Vallox sensor."""
        self._name = name
        self._state_proxy = state_proxy
        self._metric_key = metric_key
        self._device_class = device_class
        self._unit_of_measurement = unit_of_measurement
        self._icon = icon
        self._available = None
        self._state = None

    @property
    def should_poll(self):
        """Do not poll the device."""
        return False

    @property
    def name(self):
        """Return the name."""
        return self._name

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement."""
        return self._unit_of_measurement

    @property
    def device_class(self):
        """Return the device class."""
        return self._device_class

    @property
    def icon(self):
        """Return the icon."""
        return self._icon

    @property
    def available(self):
        """Return true when state is known."""
        return self._available

    @property
    def state(self):
        """Return the state."""
        return self._state

    async def async_added_to_hass(self):
        """Call to update."""
        self.async_on_remove(
            async_dispatcher_connect(
                self.hass, SIGNAL_VALLOX_STATE_UPDATE, self._update_callback
            )
        )

    @callback
    def _update_callback(self):
        """Call update method."""
        self.async_schedule_update_ha_state(True)

    async def async_update(self):
        """Fetch state from the ventilation unit."""
        try:
            self._state = self._state_proxy.fetch_metric(self._metric_key)
            self._available = True

        except (OSError, KeyError) as err:
            self._available = False
            _LOGGER.error("Error updating sensor: %s", err)


# There seems to be a quirk with respect to the fan speed reporting. The device
# keeps on reporting the last valid fan speed from when the device was in
# regular operation mode, even if it left that state and has been shut off in
# the meantime.
#
# Therefore, first query the overall state of the device, and report zero
# percent fan speed in case it is not in regular operation mode.
class ValloxFanSpeedSensor(ValloxSensor):
    """Child class for fan speed reporting."""

    async def async_update(self):
        """Fetch state from the ventilation unit."""
        try:
            # If device is in regular operation, continue.
            if self._state_proxy.fetch_metric(METRIC_KEY_MODE) == 0:
                await super().async_update()
            else:
                # Report zero percent otherwise.
                self._state = 0
                self._available = True

        except (OSError, KeyError) as err:
            self._available = False
            _LOGGER.error("Error updating sensor: %s", err)


class ValloxProfileSensor(ValloxSensor):
    """Child class for profile reporting."""

    def __init__(
        self, name, state_proxy, device_class, unit_of_measurement, icon
    ) -> None:
        """Initialize the Vallox sensor."""
        super().__init__(
            name, state_proxy, None, device_class, unit_of_measurement, icon
        )

    async def async_update(self):
        """Fetch state from the ventilation unit."""
        try:
            self._state = self._state_proxy.get_profile()
            self._available = True

        except OSError as err:
            self._available = False
            _LOGGER.error("Error updating sensor: %s", err)


class ValloxFilterRemainingSensor(ValloxSensor):
    """Child class for filter remaining time reporting."""

    async def async_update(self):
        """Fetch state from the ventilation unit."""
        try:
            days_remaining = int(self._state_proxy.fetch_metric(self._metric_key))
            days_remaining_delta = timedelta(days=days_remaining)

            # Since only a delta of days is received from the device, fix the
            # time so the timestamp does not change with every update.
            now = datetime.utcnow().replace(hour=13, minute=0, second=0, microsecond=0)

            self._state = (now + days_remaining_delta).isoformat()
            self._available = True

        except (OSError, KeyError) as err:
            self._available = False
            _LOGGER.error("Error updating sensor: %s", err)



"""Config flow for BleBox devices integration."""
import logging

from blebox_uniapi.error import Error, UnsupportedBoxVersion
from blebox_uniapi.products import Products
from blebox_uniapi.session import ApiHost
import voluptuous as vol

from homeassistant import config_entries
from homeassistant.const import CONF_HOST, CONF_PORT
from homeassistant.helpers.aiohttp_client import async_get_clientsession

from .const import (
    ADDRESS_ALREADY_CONFIGURED,
    CANNOT_CONNECT,
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_SETUP_TIMEOUT,
    DOMAIN,
    UNKNOWN,
    UNSUPPORTED_VERSION,
)

_LOGGER = logging.getLogger(__name__)


def host_port(data):
    """Return a list with host and port."""
    return (data[CONF_HOST], data[CONF_PORT])


def create_schema(previous_input=None):
    """Create a schema with given values as default."""
    if previous_input is not None:
        host, port = host_port(previous_input)
    else:
        host = DEFAULT_HOST
        port = DEFAULT_PORT

    return vol.Schema(
        {
            vol.Required(CONF_HOST, default=host): str,
            vol.Required(CONF_PORT, default=port): int,
        }
    )


LOG_MSG = {
    UNSUPPORTED_VERSION: "Outdated firmware",
    CANNOT_CONNECT: "Failed to identify device",
    UNKNOWN: "Unknown error while identifying device",
}


class BleBoxConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a config flow for BleBox devices."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_POLL

    def __init__(self):
        """Initialize the BleBox config flow."""
        self.device_config = {}

    def handle_step_exception(
        self, step, exception, schema, host, port, message_id, log_fn
    ):
        """Handle step exceptions."""

        log_fn("%s at %s:%d (%s)", LOG_MSG[message_id], host, port, exception)

        return self.async_show_form(
            step_id="user",
            data_schema=schema,
            errors={"base": message_id},
            description_placeholders={"address": f"{host}:{port}"},
        )

    async def async_step_user(self, user_input=None):
        """Handle initial user-triggered config step."""

        hass = self.hass
        schema = create_schema(user_input)

        if user_input is None:
            return self.async_show_form(
                step_id="user",
                data_schema=schema,
                errors={},
                description_placeholders={},
            )

        addr = host_port(user_input)

        for entry in hass.config_entries.async_entries(DOMAIN):
            if addr == host_port(entry.data):
                host, port = addr
                return self.async_abort(
                    reason=ADDRESS_ALREADY_CONFIGURED,
                    description_placeholders={"address": f"{host}:{port}"},
                )

        websession = async_get_clientsession(hass)
        api_host = ApiHost(*addr, DEFAULT_SETUP_TIMEOUT, websession, hass.loop, _LOGGER)

        try:
            product = await Products.async_from_host(api_host)

        except UnsupportedBoxVersion as ex:
            return self.handle_step_exception(
                "user", ex, schema, *addr, UNSUPPORTED_VERSION, _LOGGER.debug
            )

        except Error as ex:
            return self.handle_step_exception(
                "user", ex, schema, *addr, CANNOT_CONNECT, _LOGGER.warning
            )

        except RuntimeError as ex:
            return self.handle_step_exception(
                "user", ex, schema, *addr, UNKNOWN, _LOGGER.error
            )

        # Check if configured but IP changed since
        await self.async_set_unique_id(product.unique_id)
        self._abort_if_unique_id_configured()

        return self.async_create_entry(title=product.name, data=user_input)



"""Z-Wave discovery schemas."""
from . import const

DEFAULT_VALUES_SCHEMA = {
    "power": {
        const.DISC_SCHEMAS: [
            {
                const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SENSOR_MULTILEVEL],
                const.DISC_INDEX: [const.INDEX_SENSOR_MULTILEVEL_POWER],
            },
            {
                const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_METER],
                const.DISC_INDEX: [const.INDEX_METER_POWER],
            },
        ],
        const.DISC_OPTIONAL: True,
    }
}

DISCOVERY_SCHEMAS = [
    {
        const.DISC_COMPONENT: "binary_sensor",
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_ENTRY_CONTROL,
            const.GENERIC_TYPE_SENSOR_ALARM,
            const.GENERIC_TYPE_SENSOR_BINARY,
            const.GENERIC_TYPE_SWITCH_BINARY,
            const.GENERIC_TYPE_METER,
            const.GENERIC_TYPE_SENSOR_MULTILEVEL,
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_SENSOR_NOTIFICATION,
            const.GENERIC_TYPE_THERMOSTAT,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SENSOR_BINARY],
                    const.DISC_TYPE: const.TYPE_BOOL,
                    const.DISC_GENRE: const.GENRE_USER,
                },
                "off_delay": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_CONFIGURATION],
                    const.DISC_INDEX: [9],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "climate",  # thermostat without COMMAND_CLASS_THERMOSTAT_MODE
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_THERMOSTAT,
            const.GENERIC_TYPE_SENSOR_MULTILEVEL,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_THERMOSTAT_HEATING,
            const.SPECIFIC_TYPE_SETPOINT_THERMOSTAT,
            const.SPECIFIC_TYPE_NOT_USED,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_MODE]
                },
                "setpoint_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [1],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [2],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_furnace": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [7],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_dry_air": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [8],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_moist_air": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [9],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_auto_changeover": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [10],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_eco_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [11],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_eco_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [12],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_away_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [13],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_away_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [14],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_full_power": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [15],
                    const.DISC_OPTIONAL: True,
                },
                "temperature": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SENSOR_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SENSOR_MULTILEVEL_TEMPERATURE],
                    const.DISC_OPTIONAL: True,
                },
                "fan_mode": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_FAN_MODE],
                    const.DISC_OPTIONAL: True,
                },
                "operating_state": {
                    const.DISC_COMMAND_CLASS: [
                        const.COMMAND_CLASS_THERMOSTAT_OPERATING_STATE
                    ],
                    const.DISC_OPTIONAL: True,
                },
                "fan_action": {
                    const.DISC_COMMAND_CLASS: [
                        const.COMMAND_CLASS_THERMOSTAT_FAN_ACTION
                    ],
                    const.DISC_OPTIONAL: True,
                },
                "mode": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_MODE],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "climate",  # thermostat with COMMAND_CLASS_THERMOSTAT_MODE
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_THERMOSTAT,
            const.GENERIC_TYPE_SENSOR_MULTILEVEL,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_THERMOSTAT_GENERAL,
            const.SPECIFIC_TYPE_THERMOSTAT_GENERAL_V2,
            const.SPECIFIC_TYPE_SETBACK_THERMOSTAT,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_MODE]
                },
                "setpoint_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [1],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [2],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_furnace": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [7],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_dry_air": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [8],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_moist_air": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [9],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_auto_changeover": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [10],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_eco_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [11],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_eco_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [12],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_away_heating": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [13],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_away_cooling": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [14],
                    const.DISC_OPTIONAL: True,
                },
                "setpoint_full_power": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_SETPOINT],
                    const.DISC_INDEX: [15],
                    const.DISC_OPTIONAL: True,
                },
                "temperature": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SENSOR_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SENSOR_MULTILEVEL_TEMPERATURE],
                    const.DISC_OPTIONAL: True,
                },
                "fan_mode": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_THERMOSTAT_FAN_MODE],
                    const.DISC_OPTIONAL: True,
                },
                "operating_state": {
                    const.DISC_COMMAND_CLASS: [
                        const.COMMAND_CLASS_THERMOSTAT_OPERATING_STATE
                    ],
                    const.DISC_OPTIONAL: True,
                },
                "fan_action": {
                    const.DISC_COMMAND_CLASS: [
                        const.COMMAND_CLASS_THERMOSTAT_FAN_ACTION
                    ],
                    const.DISC_OPTIONAL: True,
                },
                "zxt_120_swing_mode": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_CONFIGURATION],
                    const.DISC_INDEX: [33],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "cover",  # Rollershutter
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_ENTRY_CONTROL,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_CLASS_A_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_B_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_C_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_MOTOR_MULTIPOSITION,
            const.SPECIFIC_TYPE_SECURE_BARRIER_ADDON,
            const.SPECIFIC_TYPE_SECURE_DOOR,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_GENRE: const.GENRE_USER,
                },
                "open": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SWITCH_MULTILEVEL_BRIGHT],
                    const.DISC_OPTIONAL: True,
                },
                "close": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SWITCH_MULTILEVEL_DIM],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "cover",  # Garage Door Switch
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_ENTRY_CONTROL,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_CLASS_A_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_B_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_C_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_MOTOR_MULTIPOSITION,
            const.SPECIFIC_TYPE_SECURE_BARRIER_ADDON,
            const.SPECIFIC_TYPE_SECURE_DOOR,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_BINARY],
                    const.DISC_GENRE: const.GENRE_USER,
                }
            },
        ),
    },
    {
        const.DISC_COMPONENT: "cover",  # Garage Door Barrier
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_ENTRY_CONTROL,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_CLASS_A_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_B_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_CLASS_C_MOTOR_CONTROL,
            const.SPECIFIC_TYPE_MOTOR_MULTIPOSITION,
            const.SPECIFIC_TYPE_SECURE_BARRIER_ADDON,
            const.SPECIFIC_TYPE_SECURE_DOOR,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_BARRIER_OPERATOR],
                    const.DISC_INDEX: [const.INDEX_BARRIER_OPERATOR_LABEL],
                }
            },
        ),
    },
    {
        const.DISC_COMPONENT: "fan",
        const.DISC_GENERIC_DEVICE_CLASS: [const.GENERIC_TYPE_SWITCH_MULTILEVEL],
        const.DISC_SPECIFIC_DEVICE_CLASS: [const.SPECIFIC_TYPE_FAN_SWITCH],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SWITCH_MULTILEVEL_LEVEL],
                    const.DISC_TYPE: const.TYPE_BYTE,
                }
            },
        ),
    },
    {
        const.DISC_COMPONENT: "light",
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_SWITCH_REMOTE,
        ],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_POWER_SWITCH_MULTILEVEL,
            const.SPECIFIC_TYPE_SCENE_SWITCH_MULTILEVEL,
            const.SPECIFIC_TYPE_NOT_USED,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SWITCH_MULTILEVEL_LEVEL],
                    const.DISC_TYPE: const.TYPE_BYTE,
                },
                "dimming_duration": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_MULTILEVEL],
                    const.DISC_INDEX: [const.INDEX_SWITCH_MULTILEVEL_DURATION],
                    const.DISC_OPTIONAL: True,
                },
                "color": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_COLOR],
                    const.DISC_INDEX: [const.INDEX_SWITCH_COLOR_COLOR],
                    const.DISC_OPTIONAL: True,
                },
                "color_channels": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_COLOR],
                    const.DISC_INDEX: [const.INDEX_SWITCH_COLOR_CHANNELS],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "lock",
        const.DISC_GENERIC_DEVICE_CLASS: [const.GENERIC_TYPE_ENTRY_CONTROL],
        const.DISC_SPECIFIC_DEVICE_CLASS: [
            const.SPECIFIC_TYPE_DOOR_LOCK,
            const.SPECIFIC_TYPE_ADVANCED_DOOR_LOCK,
            const.SPECIFIC_TYPE_SECURE_KEYPAD_DOOR_LOCK,
            const.SPECIFIC_TYPE_SECURE_LOCKBOX,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_DOOR_LOCK],
                    const.DISC_INDEX: [const.INDEX_DOOR_LOCK_LOCK],
                },
                "access_control": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_ALARM],
                    const.DISC_INDEX: [const.INDEX_ALARM_ACCESS_CONTROL],
                    const.DISC_OPTIONAL: True,
                },
                "alarm_type": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_ALARM],
                    const.DISC_INDEX: [const.INDEX_ALARM_TYPE],
                    const.DISC_OPTIONAL: True,
                },
                "alarm_level": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_ALARM],
                    const.DISC_INDEX: [const.INDEX_ALARM_LEVEL],
                    const.DISC_OPTIONAL: True,
                },
                "v2btze_advanced": {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_CONFIGURATION],
                    const.DISC_INDEX: [12],
                    const.DISC_OPTIONAL: True,
                },
            },
        ),
    },
    {
        const.DISC_COMPONENT: "sensor",
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [
                        const.COMMAND_CLASS_SENSOR_MULTILEVEL,
                        const.COMMAND_CLASS_METER,
                        const.COMMAND_CLASS_ALARM,
                        const.COMMAND_CLASS_SENSOR_ALARM,
                        const.COMMAND_CLASS_INDICATOR,
                        const.COMMAND_CLASS_BATTERY,
                    ],
                    const.DISC_GENRE: const.GENRE_USER,
                }
            },
        ),
    },
    {
        const.DISC_COMPONENT: "switch",
        const.DISC_GENERIC_DEVICE_CLASS: [
            const.GENERIC_TYPE_METER,
            const.GENERIC_TYPE_SENSOR_ALARM,
            const.GENERIC_TYPE_SENSOR_BINARY,
            const.GENERIC_TYPE_SWITCH_BINARY,
            const.GENERIC_TYPE_ENTRY_CONTROL,
            const.GENERIC_TYPE_SENSOR_MULTILEVEL,
            const.GENERIC_TYPE_SWITCH_MULTILEVEL,
            const.GENERIC_TYPE_SENSOR_NOTIFICATION,
            const.GENERIC_TYPE_GENERIC_CONTROLLER,
            const.GENERIC_TYPE_SWITCH_REMOTE,
            const.GENERIC_TYPE_REPEATER_SLAVE,
            const.GENERIC_TYPE_THERMOSTAT,
            const.GENERIC_TYPE_WALL_CONTROLLER,
        ],
        const.DISC_VALUES: dict(
            DEFAULT_VALUES_SCHEMA,
            **{
                const.DISC_PRIMARY: {
                    const.DISC_COMMAND_CLASS: [const.COMMAND_CLASS_SWITCH_BINARY],
                    const.DISC_TYPE: const.TYPE_BOOL,
                    const.DISC_GENRE: const.GENRE_USER,
                }
            },
        ),
    },
]



from datetime import datetime, timedelta
import logging

import homeassistant.components.automation.event as ha_event
from homeassistant.const import CONF_EVENT, CONF_ID
from homeassistant.helpers.event import async_track_time_interval

from ..pyextalife import DEVICE_ARR_ALL_TRANSMITTER
from .const import (
    CONF_EXTALIFE_EVENT_BASE,
    CONF_EXTALIFE_EVENT_TRANSMITTER,
    CONF_EXTALIFE_EVENT_UNIQUE_ID,
    CONF_PROCESSOR_EVENT_STAT_NOTIFICATION,
    CONF_PROCESSOR_EVENT_UNKNOWN,
    DOMAIN,
    EVENT_DATA,
    EVENT_TIMESTAMP,
    TRIGGER_BUTTON_DOUBLE_CLICK,
    TRIGGER_BUTTON_DOWN,
    TRIGGER_BUTTON_LONG_PRESS,
    TRIGGER_BUTTON_SINGLE_CLICK,
    TRIGGER_BUTTON_TRIPLE_CLICK,
    TRIGGER_BUTTON_UP,
    TRIGGER_SUBTYPE,
    TRIGGER_SUBTYPE_BUTTON_TEMPLATE,
    TRIGGER_TYPE,
)
from .core import Core
from .device import Device

_LOGGER = logging.getLogger(__name__)


class ExtaLifeEventProcessor:
    """ Processes status notification events from controller """

    def __init__(self, device: Device):
        self._device = device

    @staticmethod
    def factory(device: Device) -> "ExtaLifeTransmitterEventProcessor":
        if device.type in DEVICE_ARR_ALL_TRANSMITTER:
            return ExtaLifeTransmitterEventProcessor(device)

    def process_event(self, data: dict, event_type=CONF_PROCESSOR_EVENT_UNKNOWN):
        if event_type == CONF_PROCESSOR_EVENT_UNKNOWN:
            raise NotImplementedError()


class ExtaLifeTransmitterEventProcessor(ExtaLifeEventProcessor):
    def __init__(self, device: Device):
        super().__init__(device)
        self._device = device
        self._event_data = dict()
        self._event_window = dict()

    def check_supported(self, event_type):
        if event_type != CONF_PROCESSOR_EVENT_STAT_NOTIFICATION:
            raise NotImplementedError()

    def encapsulate(self, event_data) -> dict:
        event = dict()
        event[EVENT_TIMESTAMP] = datetime.now()
        event[EVENT_DATA] = event_data
        return event

    def process_event(self, data, event_type=CONF_PROCESSOR_EVENT_UNKNOWN):
        _LOGGER.debug("process_event data: %s", data)
        super().process_event(data, event_type)
        self.check_supported(event_type)

        hass = Core.get_hass()

        # assumption: data fields in JSON protocol: button & state
        button = data.get("button")
        state = data.get("state")

        event_data = {
            CONF_EXTALIFE_EVENT_UNIQUE_ID: self._device.event.unique_id,
            TRIGGER_SUBTYPE: TRIGGER_SUBTYPE_BUTTON_TEMPLATE.format(button),
        }

        if state == 1:
            event_data[CONF_TYPE] = TRIGGER_BUTTON_DOWN
        else:
            event_data[CONF_TYPE] = TRIGGER_BUTTON_UP

        def _timeout_callback(now=None):
            # assumption: state = 0 or 1
            # assumption: variable 'button' = value at the moment of callback registration
            _LOGGER.debug(
                "_timeout_callback.self._event_window[button]: %s",
                self._event_window[button],
            )
            remove_listener()
            value = ""
            for event in self._event_window[button]:
                value = value + str(event[EVENT_DATA]["state"])

            _LOGGER.debug("_timeout_callback.value: %s", value)

            event_data[TRIGGER_SUBTYPE] = TRIGGER_SUBTYPE_BUTTON_TEMPLATE.format(button)
            if value == "1":  # long press
                event_data[CONF_TYPE] = TRIGGER_BUTTON_LONG_PRESS

            elif value == "10":  # single click
                event_data[CONF_TYPE] = TRIGGER_BUTTON_SINGLE_CLICK

            elif value == "1010":  # double click
                event_data[CONF_TYPE] = TRIGGER_BUTTON_DOUBLE_CLICK

            elif value == "101010":  # triple click
                event_data[CONF_TYPE] = TRIGGER_BUTTON_TRIPLE_CLICK

            if event_data.get(TRIGGER_TYPE):
                # raise event to HA event bus
                _LOGGER.debug("_timeout_callback.async_fire event_data: %s", event_data)
                hass.bus.async_fire(self._device.event.event, event_data=event_data)

            # reset time window for button
            self._event_window[button] = None

        if self._event_window.get(button) is None:

            remove_listener = async_track_time_interval(
                hass, _timeout_callback, timedelta(milliseconds=600)
            )  # maximum tripleclick

        self._event_window.setdefault(button, []).append(self.encapsulate(data))

        if event_data.get(TRIGGER_TYPE):
            _LOGGER.debug("process_event.async_fire event_data: %s", event_data)
            hass.bus.async_fire(self._device.event.event, event_data=event_data)



"""Platform for binary sensor integration."""
import logging

from homeassistant.components.binary_sensor import BinarySensorEntity
from homeassistant.config_entries import ConfigEntry
from homeassistant.helpers.typing import HomeAssistantType

from .const import DOMAIN
from .devolo_device import DevoloDeviceEntity

_LOGGER = logging.getLogger(__name__)


async def async_setup_entry(
    hass: HomeAssistantType, entry: ConfigEntry, async_add_entities
) -> None:
    """Get all binary sensor and multi level sensor devices and setup them via config entry."""
    entities = []

    for device in hass.data[DOMAIN]["homecontrol"].binary_sensor_devices:
        for binary_sensor in device.binary_sensor_property:
            entities.append(
                DevoloBinaryDeviceEntity(
                    homecontrol=hass.data[DOMAIN]["homecontrol"],
                    device_instance=device,
                    element_uid=binary_sensor,
                )
            )

    async_add_entities(entities, False)


class DevoloBinaryDeviceEntity(DevoloDeviceEntity, BinarySensorEntity):
    """Representation of a binary sensor within devolo Home Control."""

    def __init__(self, homecontrol, device_instance, element_uid):
        """Initialize a devolo binary sensor."""
        if device_instance.binary_sensor_property.get(element_uid).sub_type != "":
            name = f"{device_instance.itemName} {device_instance.binary_sensor_property.get(element_uid).sub_type}"
        else:
            name = f"{device_instance.itemName} {device_instance.binary_sensor_property.get(element_uid).sensor_type}"

        super().__init__(
            homecontrol=homecontrol,
            device_instance=device_instance,
            element_uid=element_uid,
            name=name,
            sync=self._sync,
        )

        self._binary_sensor_property = self._device_instance.binary_sensor_property.get(
            self._unique_id
        )

        self._state = self._binary_sensor_property.state

        self._subscriber = None

    @property
    def is_on(self):
        """Return the state."""
        return self._state

    def _sync(self, message=None):
        """Update the binary sensor state."""
        if message[0].startswith("devolo.BinarySensor"):
            self._state = self._device_instance.binary_sensor_property[message[0]].state
        elif message[0].startswith("hdm"):
            self._available = self._device_instance.is_online()
        else:
            _LOGGER.debug("No valid message received: %s", message)
        self.schedule_update_ha_state()



"""Support for Acmeda Roller Blinds."""
from homeassistant.components.cover import (
    ATTR_POSITION,
    SUPPORT_CLOSE,
    SUPPORT_CLOSE_TILT,
    SUPPORT_OPEN,
    SUPPORT_OPEN_TILT,
    SUPPORT_SET_POSITION,
    SUPPORT_SET_TILT_POSITION,
    SUPPORT_STOP,
    SUPPORT_STOP_TILT,
    CoverEntity,
)
from homeassistant.core import callback
from homeassistant.helpers.dispatcher import async_dispatcher_connect

from .base import AcmedaBase
from .const import ACMEDA_HUB_UPDATE, DOMAIN
from .helpers import async_add_acmeda_entities


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up the Acmeda Rollers from a config entry."""
    hub = hass.data[DOMAIN][config_entry.entry_id]

    current = set()

    @callback
    def async_add_acmeda_covers():
        async_add_acmeda_entities(
            hass, AcmedaCover, config_entry, current, async_add_entities
        )

    hub.cleanup_callbacks.append(
        async_dispatcher_connect(
            hass,
            ACMEDA_HUB_UPDATE.format(config_entry.entry_id),
            async_add_acmeda_covers,
        )
    )


class AcmedaCover(AcmedaBase, CoverEntity):
    """Representation of a Acmeda cover device."""

    @property
    def current_cover_position(self):
        """Return the current position of the roller blind.

        None is unknown, 0 is closed, 100 is fully open.
        """
        position = None
        if self.roller.type != 7:
            position = 100 - self.roller.closed_percent
        return position

    @property
    def current_cover_tilt_position(self):
        """Return the current tilt of the roller blind.

        None is unknown, 0 is closed, 100 is fully open.
        """
        position = None
        if self.roller.type == 7 or self.roller.type == 10:
            position = 100 - self.roller.closed_percent
        return position

    @property
    def supported_features(self):
        """Flag supported features."""
        supported_features = 0
        if self.current_cover_position is not None:
            supported_features |= (
                SUPPORT_OPEN | SUPPORT_CLOSE | SUPPORT_STOP | SUPPORT_SET_POSITION
            )
        if self.current_cover_tilt_position is not None:
            supported_features |= (
                SUPPORT_OPEN_TILT
                | SUPPORT_CLOSE_TILT
                | SUPPORT_STOP_TILT
                | SUPPORT_SET_TILT_POSITION
            )

        return supported_features

    @property
    def is_closed(self):
        """Return if the cover is closed."""
        is_closed = self.roller.closed_percent == 100
        return is_closed

    async def close_cover(self, **kwargs):
        """Close the roller."""
        await self.roller.move_down()

    async def open_cover(self, **kwargs):
        """Open the roller."""
        await self.roller.move_up()

    async def stop_cover(self, **kwargs):
        """Stop the roller."""
        await self.roller.move_stop()

    async def set_cover_position(self, **kwargs):
        """Move the roller shutter to a specific position."""
        await self.roller.move_to(100 - kwargs[ATTR_POSITION])

    async def close_cover_tilt(self, **kwargs):
        """Close the roller."""
        await self.roller.move_down()

    async def open_cover_tilt(self, **kwargs):
        """Open the roller."""
        await self.roller.move_up()

    async def stop_cover_tilt(self, **kwargs):
        """Stop the roller."""
        await self.roller.move_stop()

    async def set_cover_tilt(self, **kwargs):
        """Tilt the roller shutter to a specific position."""
        await self.roller.move_to(100 - kwargs[ATTR_POSITION])
